import json
import urllib.request
from enum import StrEnum

import polars as pl

from spells.enums import ColName


class CardAttr(StrEnum):
    NAME = ColName.NAME
    SET_CODE = ColName.SET_CODE
    COLOR = ColName.COLOR
    RARITY = ColName.RARITY
    COLOR_IDENTITY = ColName.COLOR_IDENTITY
    CARD_TYPE = ColName.CARD_TYPE
    SUBTYPE = ColName.SUBTYPE
    MANA_VALUE = ColName.MANA_VALUE
    MANA_COST = ColName.MANA_COST
    POWER = ColName.POWER
    TOUGHNESS = ColName.TOUGHNESS
    IS_BONUS_SHEET = ColName.IS_BONUS_SHEET
    IS_DFC = ColName.IS_DFC
    ORACLE_TEXT = ColName.ORACLE_TEXT
    CARD_JSON = ColName.CARD_JSON


MTG_JSON_TEMPLATE = "https://mtgjson.com/api/v5/{set_code}.json"


def _fetch_mtg_json(set_code: str) -> dict:
    request = urllib.request.Request(
        MTG_JSON_TEMPLATE.format(set_code=set_code),
        headers={"User-Agent": "spells-mtg/0.1.0"},
    )

    with urllib.request.urlopen(request) as f:
        draft_set_json = json.loads(f.read().decode("utf-8"))

    return draft_set_json


def _extract_value(set_code: str, name: str, card_dict: dict, field: CardAttr):
    match field:
        case CardAttr.NAME:
            return name
        case CardAttr.SET_CODE:
            return card_dict.get("setCode", "")
        case CardAttr.COLOR:
            return "".join(card_dict.get("colors", []))
        case CardAttr.RARITY:
            return card_dict.get("rarity", "")
        case CardAttr.COLOR_IDENTITY:
            return "".join(card_dict.get("colorIdentity", []))
        case CardAttr.CARD_TYPE:
            return " ".join(card_dict.get("types", []))
        case CardAttr.SUBTYPE:
            return " ".join(card_dict.get("subtypes", []))
        case CardAttr.MANA_VALUE:
            return card_dict.get("manaValue", 0)
        case CardAttr.MANA_COST:
            return card_dict.get("manaCost", "")
        case CardAttr.POWER:
            return card_dict.get("power", None)
        case CardAttr.TOUGHNESS:
            return card_dict.get("toughness", None)
        case CardAttr.IS_BONUS_SHEET:
            return card_dict.get("setCode", set_code) != set_code
        case CardAttr.IS_DFC:
            return len(card_dict.get("otherFaceIds", [])) > 0
        case CardAttr.ORACLE_TEXT:
            return card_dict.get("text", "")
        case CardAttr.CARD_JSON:
            return card_dict.get("json", "")


def card_df(draft_set_code: str, names: list[str]) -> pl.DataFrame:
    draft_set_json = _fetch_mtg_json(draft_set_code)
    booster_info = draft_set_json["data"]["booster"]

    booster_type = "play" if "play" in booster_info else "draft"
    set_codes = booster_info[booster_type]["sourceSetCodes"]
    set_codes.reverse()

    card_data_map = {}
    for set_code in set_codes:
        if set_code != draft_set_code:
            card_data = _fetch_mtg_json(set_code)["data"]["cards"]
        else:
            card_data = draft_set_json["data"]["cards"]

        card_data.reverse()  # prefer front face for split cards
        for item in card_data:
            item["json"] = json.dumps(item)

        face_name_cards = [item for item in card_data if "faceName" in item]
        card_data_map.update({item["faceName"]: item for item in face_name_cards})
        card_data_map.update({item["name"]: item for item in card_data})

    return pl.DataFrame(
        [
            {
                field: _extract_value(
                    draft_set_code, name, card_data_map.get(name, {}), field
                )
                for field in CardAttr
            }
            for name in names
        ]
    )



# Downloaded from 17Lands.com on 2024-10-17 under below license
# https://www.17lands.com/public_datasets
# https://creativecommons.org/licenses/by/4.0/
#
# Modified by :s%/float16/float64/g
# converted to use by polars

import csv
import re
from typing import Dict

import polars as pl

COLUMN_TYPES = (
    # Metadata
    (re.compile(r"^draft_id_idx$"), pl.Int64),
    (re.compile(r"^user_n_games_bucket$"), pl.Int16),
    (re.compile(r"^user_game_win_rate_bucket$"), pl.Float64),
    (re.compile(r"^expansion$"), pl.String),
    (re.compile(r"^event_type$"), pl.String),
    (re.compile(r"^draft_id$"), pl.String),
    (re.compile(r"^draft_time$"), pl.String),
    (re.compile(r"^rank$"), pl.String),
    # Draft
    (re.compile(r"^event_match_wins$"), pl.Int8),
    (re.compile(r"^event_match_losses$"), pl.Int8),
    (re.compile(r"^pack_number$"), pl.Int8),
    (re.compile(r"^pick_number$"), pl.Int8),
    (re.compile(r"^pick$"), pl.String),
    (re.compile(r"^pick_maindeck_rate$"), pl.Float64),
    (re.compile(r"^pick_sideboard_in_rate$"), pl.Float64),
    (re.compile(r"^pool_.*"), pl.Int8),
    (re.compile(r"^pack_card_.*"), pl.Int8),
    # Game + Replay
    (re.compile(r"^game_time$"), pl.String),
    (re.compile(r"^build_index$"), pl.Int8),
    (re.compile(r"^match_number$"), pl.Int8),
    (re.compile(r"^game_number$"), pl.Int8),
    (re.compile(r"^opp_rank$"), pl.String),
    (re.compile(r"^main_colors$"), pl.String),
    (re.compile(r"^splash_colors$"), pl.String),
    (re.compile(r"^on_play$"), pl.Boolean),
    (re.compile(r"^num_mulligans$"), pl.Int8),
    (re.compile(r"^opp_num_mulligans$"), pl.Int8),
    (re.compile(r"^opp_colors$"), pl.String),
    (re.compile(r"^num_turns$"), pl.Int8),
    (re.compile(r"^won$"), pl.Boolean),
    (re.compile(r"^deck_.*"), pl.Int8),
    (re.compile(r"^sideboard_.*"), pl.Int8),
    # Game
    (re.compile(r"^drawn_.*"), pl.Int8),
    (re.compile(r"^tutored_.*"), pl.Int8),
    (re.compile(r"^opening_hand_.*"), pl.Int8),
    # Replay
    (re.compile(r"^candidate_hand_\d$"), pl.String),
    (re.compile(r"^opening_hand$"), pl.String),
    (re.compile(r"^user_turn_\d+_cards_drawn$"), pl.String),
    (re.compile(r"^((user)|(oppo))_turn_\d+_cards_discarded$"), pl.String),
    (re.compile(r"^((user)|(oppo))_turn_\d+_lands_played$"), pl.String),
    (re.compile(r"^user_turn_\d+_cards_foretold$"), pl.String),
    (re.compile(r"^((user)|(oppo))_turn_\d+_creatures_cast$"), pl.String),
    (re.compile(r"^((user)|(oppo))_turn_\d+_non_creatures_cast$"), pl.String),
    (
        re.compile(
            r"^((user)|(oppo))_turn_\d+_((user)|(oppo))_instants_sorceries_cast$"
        ),
        pl.String,
    ),
    (re.compile(r"^((user)|(oppo))_turn_\d+_((user)|(oppo))_abilities$"), pl.String),
    (
        re.compile(r"^((user)|(oppo))_turn_\d+_((user)|(oppo))_cards_learned$"),
        pl.String,
    ),
    (re.compile(r"^((user)|(oppo))_turn_\d+_creatures_attacked$"), pl.String),
    (re.compile(r"^((user)|(oppo))_turn_\d+_creatures_blocked$"), pl.String),
    (re.compile(r"^((user)|(oppo))_turn_\d+_creatures_unblocked$"), pl.String),
    (re.compile(r"^((user)|(oppo))_turn_\d+_creatures_blocking$"), pl.String),
    (re.compile(r"^((user)|(oppo))_turn_\d+_creatures_blitzed$"), pl.Int8),
    (
        re.compile(r"^((user)|(oppo))_turn_\d+_player_combat_damage_dealt$"),
        pl.String,
    ),  # DEPRECATED
    (
        re.compile(r"^((user)|(oppo))_turn_\d+_((user)|(oppo))_combat_damage_taken$"),
        pl.String,
    ),
    (
        re.compile(
            r"^((user)|(oppo))_turn_\d+_((user)|(oppo))_creatures_killed_combat$"
        ),
        pl.String,
    ),
    (
        re.compile(
            r"^((user)|(oppo))_turn_\d+_((user)|(oppo))_creatures_killed_non_combat$"
        ),
        pl.String,
    ),
    (re.compile(r"^((user)|(oppo))_turn_\d+_((user)|(oppo))_mana_spent$"), pl.Float64),
    (re.compile(r"^((user)|(oppo))_turn_\d+_eot_user_cards_in_hand$"), pl.String),
    (re.compile(r"^((user)|(oppo))_turn_\d+_eot_oppo_cards_in_hand$"), pl.Float64),
    (
        re.compile(r"^((user)|(oppo))_turn_\d+_eot_((user)|(oppo))_lands_in_play$"),
        pl.String,
    ),
    (
        re.compile(r"^((user)|(oppo))_turn_\d+_eot_((user)|(oppo))_creatures_in_play$"),
        pl.String,
    ),
    (
        re.compile(
            r"^((user)|(oppo))_turn_\d+_eot_((user)|(oppo))_non_creatures_in_play$"
        ),
        pl.String,
    ),
    (re.compile(r"^((user)|(oppo))_turn_\d+_eot_((user)|(oppo))_life$"), pl.Float64),
    (
        re.compile(r"^((user)|(oppo))_turn_\d+_eot_((user)|(oppo))_poison_counters$"),
        pl.Float64,
    ),
    (re.compile(r"^user_turn_\d+_cards_tutored$"), pl.String),
    (re.compile(r"^oppo_turn_\d+_cards_tutored$"), pl.Int8),
    (re.compile(r"^oppo_turn_\d+_cards_drawn_or_tutored$"), pl.Int8),
    (re.compile(r"^oppo_turn_\d+_cards_drawn$"), pl.Int8),
    (re.compile(r"^oppo_turn_\d+_cards_foretold$"), pl.Int8),
    (re.compile(r"^((user)|(oppo))_total_cards_drawn$"), pl.Int8),
    (re.compile(r"^((user)|(oppo))_total_cards_discarded$"), pl.Int8),
    (re.compile(r"^((user)|(oppo))_total_lands_played$"), pl.Int8),
    (re.compile(r"^((user)|(oppo))_total_cards_foretold$"), pl.Int8),
    (re.compile(r"^((user)|(oppo))_total_creatures_cast$"), pl.Int8),
    (re.compile(r"^((user)|(oppo))_total_creatures_blitzed$"), pl.Int8),
    (re.compile(r"^((user)|(oppo))_total_non_creatures_cast$"), pl.Int8),
    (re.compile(r"^((user)|(oppo))_total_instants_sorceries_cast$"), pl.Int8),
    (re.compile(r"^((user)|(oppo))_total_cards_learned$"), pl.Int8),
    (re.compile(r"^((user)|(oppo))_total_mana_spent$"), pl.Int16),
    (re.compile(r"^oppo_total_cards_drawn_or_tutored$"), pl.Int8),
)


def schema(
    filename: str, print_missing: bool = False
) -> Dict[str, pl.datatypes.DataType]:
    dtypes: Dict[str, pl.datatypes.DataType] = {}
    with open(filename, encoding="utf-8") as f:
        columns = csv.DictReader(f).fieldnames
    if columns is None:
        raise ValueError(f"Could not read fieldnames from {filename}")
    for column in columns:
        for regex, column_type in COLUMN_TYPES:
            if regex.match(column):
                dtypes[column] = column_type
                break
            else:
                if print_missing:
                    print(f"Could not find an appropriate type for {column}")
    return dtypes



"""
this is where calculations are performed on the 17Lands public data sets and
aggregate calculations are returned.

Aggregate dataframes containing raw counts are cached in the local file system
for performance.
"""

import functools
import hashlib
import re
from inspect import signature
import os
from typing import Callable, TypeVar, Any

import polars as pl
from polars.exceptions import ColumnNotFoundError

from spells.external import data_file_path
import spells.cache
import spells.filter
import spells.manifest
from spells.columns import ColDef, ColSpec, get_specs
from spells.enums import View, ColName, ColType


DF = TypeVar("DF", pl.LazyFrame, pl.DataFrame)


def _cache_key(args) -> str:
    """
    cache arguments by __str__ (based on the current value of a mutable, so be careful)
    """
    return hashlib.md5(str(args).encode("utf-8")).hexdigest()


@functools.lru_cache(maxsize=None)
def _get_names(set_code: str) -> list[str]:
    card_fp = data_file_path(set_code, View.CARD)
    card_view = pl.read_parquet(card_fp)
    card_names_set = frozenset(card_view.get_column("name").to_list())

    draft_fp = data_file_path(set_code, View.DRAFT)
    draft_view = pl.scan_parquet(draft_fp)
    cols = draft_view.collect_schema().names()

    prefix = "pack_card_"
    names = [col[len(prefix) :] for col in cols if col.startswith(prefix)]
    draft_names_set = frozenset(names)

    assert (
        draft_names_set == card_names_set
    ), "names mismatch between card and draft file"
    return names


def _get_card_context(
    set_code: str,
    specs: dict[str, ColSpec],
    card_context: pl.DataFrame | dict[str, dict[str, Any]] | None,
    set_context: pl.DataFrame | dict[str, Any] | None,
    card_only: bool = False,
) -> dict[str, dict[str, Any]]:
    card_attr_specs = {
        col: spec
        for col, spec in specs.items()
        if spec.col_type == ColType.CARD_ATTR or col == ColName.NAME
    }

    if not card_only:
        col_def_map = _hydrate_col_defs(
            set_code,
            card_attr_specs,
            set_context=set_context,
            card_context=card_context,
            card_only=True,
        )

        columns = list(col_def_map.keys())

        fp = data_file_path(set_code, View.CARD)
        card_df = pl.read_parquet(fp)
        select_rows = _view_select(
            card_df, frozenset(columns), col_def_map, is_agg_view=False
        ).to_dicts()

        loaded_context = {row[ColName.NAME]: row for row in select_rows}
    else:
        names = _get_names(set_code)
        loaded_context = {name: {} for name in names}

    if card_context is not None:
        if isinstance(card_context, pl.DataFrame):
            try:
                card_context = {
                    row[ColName.NAME]: row for row in card_context.to_dicts()
                }
            except ColumnNotFoundError:
                raise ValueError("card_context DataFrame must have column 'name'")

        names = list(loaded_context.keys())
        for name in names:
            assert (
                name in card_context
            ), f"card_context must include a row for each card name. {name} missing."
            for col, value in card_context[name].items():
                loaded_context[name][col] = value

    return loaded_context


def _determine_expression(
    col: str,
    spec: ColSpec,
    names: list[str],
    card_context: dict[str, dict],
    set_context: dict[str, Any],
) -> pl.Expr | tuple[pl.Expr, ...]:
    def seed_params(expr):
        params = {}

        sig_params = signature(expr).parameters
        if "names" in sig_params:
            params["names"] = names
        if "card_context" in sig_params:
            params["card_context"] = card_context
        if "set_context" in sig_params:
            params["set_context"] = set_context
        return params

    if spec.col_type == ColType.NAME_SUM:
        if spec.expr is not None:
            assert isinstance(
                spec.expr, Callable
            ), f"NAME_SUM column {col} must have a callable `expr` accepting a `name` argument"
            unnamed_exprs = [
                spec.expr(**{"name": name, **seed_params(spec.expr)}) for name in names
            ]

            expr = tuple(
                map(
                    lambda ex, name: ex.alias(f"{col}_{name}"),
                    unnamed_exprs,
                    names,
                )
            )
        else:
            expr = tuple(map(lambda name: pl.col(f"{col}_{name}"), names))

    elif spec.expr is not None:
        if isinstance(spec.expr, Callable):
            assert (
                not spec.col_type == ColType.AGG
            ), f"AGG column {col} must be a pure spells expression"
            params = seed_params(spec.expr)
            if (
                spec.col_type in (ColType.PICK_SUM, ColType.CARD_ATTR)
                and "name" in signature(spec.expr).parameters
            ):
                condition_col = (
                    ColName.PICK if spec.col_type == ColType.PICK_SUM else ColName.NAME
                )
                expr = pl.lit(None)
                for name in names:
                    name_params = {"name": name, **params}
                    expr = (
                        pl.when(pl.col(condition_col) == name)
                        .then(spec.expr(**name_params))
                        .otherwise(expr)
                    )
            else:
                expr = spec.expr(**params)
        else:
            expr = spec.expr
        expr = expr.alias(col)

        if spec.col_type == ColType.AGG:
            expr = expr.fill_nan(None)
    else:
        expr = pl.col(col)

    return expr


def _infer_dependencies(
    name: str,
    expr: pl.Expr | tuple[pl.Expr, ...],
    specs: dict[str, ColSpec],
    names: list[str],
) -> set[str]:
    dependencies = set()
    tricky_ones = set()

    if isinstance(expr, pl.Expr):
        dep_cols = [c for c in expr.meta.root_names() if c != name]
        for dep_col in dep_cols:
            if dep_col in specs.keys():
                dependencies.add(dep_col)
            else:
                tricky_ones.add(dep_col)
    else:
        for idx, exp in enumerate(expr):
            pattern = f"_{names[idx]}$"
            dep_cols = [c for c in exp.meta.root_names() if c != name]
            for dep_col in dep_cols:
                if dep_col in specs.keys():
                    dependencies.add(dep_col)
                elif (
                    len(split := re.split(pattern, dep_col)) == 2 and split[0] in specs
                ):
                    dependencies.add(split[0])
                else:
                    tricky_ones.add(dep_col)

    for item in tricky_ones:
        found = False
        for n in names:
            pattern = f"_{n}$"
            if (
                not found
                and len(split := re.split(pattern, item)) == 2
                and split[0] in specs
            ):
                dependencies.add(split[0])
                found = True
        assert found, f"Could not locate column spec for root col {item}"

    return dependencies


def _get_set_context(
    set_code: str, set_context: pl.DataFrame | dict[str, Any] | None
) -> dict[str, Any]:
    context_fp = data_file_path(set_code, "context")

    report = functools.partial(
        spells.cache.spells_print,
        "report",
        f"Set context for {set_code} invalid, please investigate!",
    )

    context = {}
    if not os.path.isfile(context_fp):
        report()
    else:
        context_df = pl.read_parquet(context_fp)
        if len(context_df) == 1:
            context.update(context_df.to_dicts()[0])
        else:
            report()

    if isinstance(set_context, pl.DataFrame):
        assert len(set_context != 1), "Invalid set context provided"
        context.update(set_context.to_dicts()[0])
    elif isinstance(set_context, dict):
        context.update(set_context)

    return context


def _hydrate_col_defs(
    set_code: str,
    specs: dict[str, ColSpec],
    card_context: pl.DataFrame | dict[str, dict] | None = None,
    set_context: pl.DataFrame | dict[str, Any] | None = None,
    card_only: bool = False,
):
    names = _get_names(set_code)

    set_context = _get_set_context(set_code, set_context)

    card_context = _get_card_context(
        set_code, specs, card_context, set_context, card_only=card_only
    )

    assert len(names) > 0, "there should be names"
    hydrated = {}
    for col, spec in specs.items():
        expr = _determine_expression(col, spec, names, card_context, set_context)
        dependencies = _infer_dependencies(col, expr, specs, names)

        sig_expr = expr if isinstance(expr, pl.Expr) else expr[0]
        try:
            expr_sig = sig_expr.meta.serialize(format="json")
        except pl.exceptions.ComputeError:
            if spec.version is not None:
                expr_sig = col + spec.version
            else:
                print(
                    f"Using session-only signature for non-serializable column {col}, please provide a version value"
                )
                expr_sig = str(sig_expr)

        signature = str(
            (
                col,
                spec.col_type.value,
                expr_sig,
                sorted(dependencies),
            )
        )

        cdef = ColDef(
            name=col,
            col_type=spec.col_type,
            views=set(spec.views or set()),
            expr=expr,
            dependencies=dependencies,
            signature=signature,
        )
        hydrated[col] = cdef
    return hydrated


def _view_select(
    df: DF,
    view_cols: frozenset[str],
    col_def_map: dict[str, ColDef],
    is_agg_view: bool,
) -> DF:
    base_cols = frozenset()
    cdefs = [col_def_map[c] for c in sorted(view_cols)]
    select = []
    for cdef in cdefs:
        if is_agg_view:
            if cdef.col_type == ColType.AGG:
                base_cols = base_cols.union(cdef.dependencies)
                select.append(cdef.expr)
            else:
                base_cols = base_cols.union(frozenset({cdef.name}))
                select.append(cdef.name)
        else:
            if cdef.dependencies:
                base_cols = base_cols.union(cdef.dependencies)
            else:
                base_cols = base_cols.union(frozenset({cdef.name}))
            if isinstance(cdef.expr, tuple):
                select.extend(cdef.expr)
            else:
                select.append(cdef.expr)

    if base_cols != view_cols:
        df = _view_select(df, base_cols, col_def_map, is_agg_view)

    return df.select(select)


def _fetch_or_cache(
    calc_fn: Callable,
    set_code: str,
    cache_args,
    read_cache: bool = True,
    write_cache: bool = True,
):
    key = _cache_key(cache_args)

    if read_cache:
        if spells.cache.cache_exists(set_code, key):
            return spells.cache.read_cache(set_code, key)

    df = calc_fn()

    if write_cache:
        spells.cache.write_cache(set_code, key, df)

    return df


def _base_agg_df(
    set_code: str,
    m: spells.manifest.Manifest,
    use_streaming: bool = False,
) -> pl.DataFrame:
    join_dfs = []
    group_by = m.base_view_group_by

    is_name_gb = ColName.NAME in group_by
    nonname_gb = tuple(gb for gb in group_by if gb != ColName.NAME)

    for view, cols_for_view in m.view_cols.items():
        if view == View.CARD:
            continue
        df_path = data_file_path(set_code, view)
        base_view_df = pl.scan_parquet(df_path)
        base_df_prefilter = _view_select(
            base_view_df, cols_for_view, m.col_def_map, is_agg_view=False
        )

        if m.filter is not None:
            base_df = base_df_prefilter.filter(m.filter.expr)
        else:
            base_df = base_df_prefilter

        sum_cols = tuple(
            c
            for c in cols_for_view
            if m.col_def_map[c].col_type in (ColType.PICK_SUM, ColType.GAME_SUM)
        )
        if sum_cols:
            # manifest will verify that GAME_SUM manifests do not use NAME grouping
            name_col_tuple = (
                (pl.col(ColName.PICK).alias(ColName.NAME),) if is_name_gb else ()
            )

            sum_col_df = base_df.select(nonname_gb + name_col_tuple + sum_cols)

            grouped = sum_col_df.group_by(group_by) if group_by else sum_col_df
            join_dfs.append(grouped.sum().collect(streaming=use_streaming))

        name_sum_cols = tuple(
            c for c in cols_for_view if m.col_def_map[c].col_type == ColType.NAME_SUM
        )
        for col in name_sum_cols:
            cdef = m.col_def_map[col]
            pattern = f"^{cdef.name}_"
            name_map = functools.partial(
                lambda patt, name: re.split(patt, name)[1], pattern
            )

            expr = pl.col(f"^{cdef.name}_.*$").name.map(name_map)
            pre_agg_df = base_df.select((expr,) + nonname_gb)

            if nonname_gb:
                agg_df = pre_agg_df.group_by(nonname_gb).sum()
            else:
                agg_df = pre_agg_df.sum()

            index = nonname_gb if nonname_gb else None
            unpivoted = agg_df.unpivot(
                index=index,
                value_name=m.col_def_map[col].name,
                variable_name=ColName.NAME,
            )

            if not is_name_gb:
                grouped = (
                    unpivoted.drop("name").group_by(nonname_gb)
                    if nonname_gb
                    else unpivoted.drop("name")
                )
                df = grouped.sum().collect(streaming=use_streaming)
            else:
                df = unpivoted.collect(streaming=use_streaming)

            join_dfs.append(df)

    if group_by:
        joined_df = functools.reduce(
            lambda prev, curr: prev.join(curr, on=group_by, how="outer", coalesce=True),
            join_dfs,
        )
    else:
        joined_df = pl.concat(join_dfs, how="horizontal")

    return joined_df


def summon(
    set_code: str | list[str],
    columns: list[str] | None = None,
    group_by: list[str] | None = None,
    filter_spec: dict | None = None,
    extensions: dict[str, ColSpec] | list[dict[str, ColSpec]] | None = None,
    use_streaming: bool = False,
    read_cache: bool = True,
    write_cache: bool = True,
    card_context: pl.DataFrame | dict[str, Any] | None = None,
    set_context: pl.DataFrame | dict[str, Any] | None = None,
) -> pl.DataFrame:
    specs = get_specs()

    if extensions is not None:
        if not isinstance(extensions, list):
            extensions = [extensions]
        for ext in extensions:
            specs.update(ext)

    if isinstance(set_code, str):
        card_context = {set_code: card_context}
        set_context = {set_code: set_context}
        codes = [set_code]
    else:
        codes = set_code

    assert codes, "Please ask for at least one set"

    m = None

    concat_dfs = []
    for code in codes:
        if isinstance(card_context, pl.DataFrame):
            set_card_context = card_context.filter(pl.col("expansion") == code)
        elif isinstance(card_context, dict):
            set_card_context = card_context[code]
        else:
            set_card_context = None

        if isinstance(set_context, pl.DataFrame):
            this_set_context = set_context.filter(pl.col("expansion") == code)
        elif isinstance(set_context, dict):
            this_set_context = set_context[code]
        else:
            this_set_context = None

        col_def_map = _hydrate_col_defs(code, specs, set_card_context, this_set_context)
        m = spells.manifest.create(col_def_map, columns, group_by, filter_spec)

        calc_fn = functools.partial(_base_agg_df, code, m, use_streaming=use_streaming)
        agg_df = _fetch_or_cache(
            calc_fn,
            code,
            (
                code,
                sorted(m.view_cols.get(View.DRAFT, set())),
                sorted(m.view_cols.get(View.GAME, set())),
                sorted(c.signature or "" for c in m.col_def_map.values()),
                sorted(m.base_view_group_by),
                filter_spec,
            ),
            read_cache=read_cache,
            write_cache=write_cache,
        )

        if View.CARD in m.view_cols:
            card_cols = m.view_cols[View.CARD].union({ColName.NAME})
            fp = data_file_path(code, View.CARD)
            card_df = pl.read_parquet(fp)
            select_df = _view_select(
                card_df, card_cols, m.col_def_map, is_agg_view=False
            )
            agg_df = agg_df.join(select_df, on="name", how="outer", coalesce=True)
        concat_dfs.append(agg_df)

    full_agg_df = pl.concat(concat_dfs, how="vertical")

    assert (
        m is not None
    ), "What happened? We mean to use one of the sets manifest, it shouldn't matter which."

    if m.group_by:
        gb = m.group_by
        # an agg may depend on some card column that hasn't been explicitly requested, but can be safely
        # depended on if we are grouping by name
        if ColName.NAME in m.group_by and View.CARD in m.view_cols:
            for col in m.view_cols[View.CARD]:
                if col not in m.group_by:
                    gb = tuple([*gb, col])
        full_agg_df = full_agg_df.group_by(gb).sum()
    else:
        full_agg_df = full_agg_df.sum()

    ret_cols = m.group_by + m.columns
    ret_df = (
        _view_select(full_agg_df, frozenset(ret_cols), m.col_def_map, is_agg_view=True)
        .select(ret_cols)
        .sort(m.group_by)
    )

    return ret_df



"""
spells.filter (don't import as builtin filter) returns a function from_spec
that takes a dict-specified filter and returns a Filter object that records
the dependent column names and contains a filter expression for use in polars.
"""

from dataclasses import dataclass
import functools

import polars as pl


@dataclass(frozen=True)
class Filter:
    expr: pl.Expr
    lhs: frozenset[str]


def __negate(f: Filter) -> Filter:
    return Filter(expr=~f.expr, lhs=f.lhs)


def _or(f1: Filter, f2: Filter) -> Filter:
    return Filter(expr=f1.expr | f2.expr, lhs=f1.lhs.union(f2.lhs))


def _and(f1: Filter, f2: Filter) -> Filter:
    return Filter(expr=f1.expr & f2.expr, lhs=f1.lhs.union(f2.lhs))


def _filter_eq(lhs: str, rhs: str) -> Filter:
    return Filter(expr=pl.col(lhs) == rhs, lhs=frozenset({lhs}))


def _filter_leq(lhs: str, rhs: str) -> Filter:
    return Filter(expr=pl.col(lhs) <= rhs, lhs=frozenset({lhs}))


def _filter_geq(lhs: str, rhs: str) -> Filter:
    return Filter(expr=pl.col(lhs) >= rhs, lhs=frozenset({lhs}))


def _filter_in(lhs: str, rhs: str) -> Filter:
    return Filter(expr=pl.col(lhs).is_in(rhs), lhs=frozenset({lhs}))


def _filter_gt(lhs: str, rhs: str) -> Filter:
    return __negate(_filter_leq(lhs, rhs))


def _filter_nin(lhs: str, rhs: str) -> Filter:
    return __negate(_filter_in(lhs, rhs))


def _filter_neq(lhs: str, rhs: str) -> Filter:
    return __negate(_filter_eq(lhs, rhs))


def _filter_lt(lhs: str, rhs: str) -> Filter:
    return __negate(_filter_geq(lhs, rhs))


filter_fn_map = {
    "=": _filter_eq,
    "<=": _filter_leq,
    ">=": _filter_geq,
    "in": _filter_in,
    ">": _filter_gt,
    "nin": _filter_nin,
    "!=": _filter_neq,
    "<": _filter_lt,
}


def _base(lhs, rhs, op="=") -> Filter:
    return filter_fn_map[op](lhs, rhs)


def _all_of(filters) -> Filter:
    return functools.reduce(_and, filters)


def _any_of(filters) -> Filter:
    return functools.reduce(_or, filters)


def _negate(fil) -> Filter:
    return __negate(fil)


BUILDER_MAP = {"$and": _all_of, "$or": _any_of, "$not": _negate}


def from_spec(filter_spec: dict | None) -> Filter | None:
    """
    filter_spec is a nested dictionary with the leaf-level consisting of specs of the form
    {'lhs': 'a', 'rhs': [1,2,3], 'op': 'in'}
    or
    {'a': 5}

    higher level keys can be `_all_of | _any_of | not`

    e.g.

    {
        '$and': [
            {
                '$not': {'A' : 5}
            },
            {
                {'lhs': 'B', 'rhs': [1,2], 'op': 'in'}
            }
        ]
    }

    an empty input returns None, which represents a trivial filter
    """
    if not filter_spec:
        return None

    for filter_type, filter_fn in BUILDER_MAP.items():
        if filter_value := filter_spec.get(filter_type):
            assert (
                len(filter_spec) == 1
            ), f"Operator {filter_type} incompatible with additional keys."
            if isinstance(filter_value, list):
                arg = tuple(map(from_spec, filter_value))
            else:
                arg = from_spec(filter_value)
            return filter_fn(arg)

    if len(filter_spec) == 1:
        for lhs, rhs in filter_spec.items():
            return _base(lhs, rhs)

    assert "lhs" in filter_spec and "rhs" in filter_spec
    return _base(filter_spec["lhs"], filter_spec["rhs"], filter_spec.get("op", "="))



"""
enums
"""

from enum import StrEnum


class View(StrEnum):
    GAME = "game"
    DRAFT = "draft"
    CARD = "card"


class ColType(StrEnum):
    FILTER_ONLY = "filter_only"
    GROUP_BY = "group_by"
    PICK_SUM = "pick_sum"
    GAME_SUM = "game_sum"
    NAME_SUM = "name_sum"
    AGG = "agg"
    CARD_ATTR = "card_attr"


class ColName(StrEnum):
    """
    A list of all available columns, including built-in extensions.

    "Name-mapped" columns like "deck_<card name>" are identified by the prefix only.
    Those columns can be referenced simply as e.g. "deck" in formulas for the post-agg stage.

    The definitions of the columns and how they may be used is defined in `column_defs`
    """

    # shared
    NAME = "name"  # special column for card name index
    EXPANSION = "expansion"
    EVENT_TYPE = "event_type"
    DRAFT_ID = "draft_id"
    DRAFT_TIME = "draft_time"  # modified, cast to time
    DRAFT_DATE = "draft_date"
    FORMAT_DAY = "format_day"
    DRAFT_DAY_OF_WEEK = "draft_day_of_week"
    DRAFT_HOUR = "draft_hour"
    DRAFT_WEEK = "draft_week"
    FORMAT_WEEK = "format_week"
    RANK = "rank"
    USER_N_GAMES_BUCKET = "user_n_games_bucket"
    USER_GAME_WIN_RATE_BUCKET = "user_game_win_rate_bucket"
    PLAYER_COHORT = "player_cohort"
    # draft
    EVENT_MATCH_WINS = "event_match_wins"
    EVENT_MATCH_WINS_SUM = "event_match_wins_sum"
    EVENT_MATCH_LOSSES = "event_match_losses"
    EVENT_MATCH_LOSSES_SUM = "event_match_losses_sum"
    EVENT_MATCHES = "event_matches"
    EVENT_MATCHES_SUM = "event_matches_sum"
    IS_TROPHY = "is_trophy"
    IS_TROPHY_SUM = "is_trophy_sum"
    PACK_NUMBER = "pack_number"
    PACK_NUM = "pack_num"  # pack_number plus 1
    PICK_NUMBER = "pick_number"
    PICK_NUM = "pick_num"  # pick_number plus 1
    TAKEN_AT = "taken_at"
    NUM_TAKEN = "num_taken"
    NUM_DRAFTS = "num_drafts"
    PICK = "pick"
    PICK_MAINDECK_RATE = "pick_maindeck_rate"
    PICK_SIDEBOARD_IN_RATE = "pick_sideboard_in_rate"
    PACK_CARD = "pack_card"
    LAST_SEEN = "last_seen"
    NUM_SEEN = "num_seen"
    POOL = "pool"
    # game
    GAME_TIME = "game_time"
    GAME_DATE = "game_date"
    GAME_DAY_OF_WEEK = "game_day_of_week"
    GAME_HOUR = "game_hour"
    GAME_WEEK = "game_week"
    BUILD_INDEX = "build_index"
    MATCH_NUMBER = "match_number"
    GAME_NUMBER = "game_number"
    NUM_EVENTS = "num_events"
    NUM_MATCHES = "num_matches"
    NUM_GAMES = "num_games"
    OPP_RANK = "opp_rank"  # not populated for recent sets
    MAIN_COLORS = "main_colors"
    NUM_COLORS = "num_colors"
    SPLASH_COLORS = "splash_colors"
    HAS_SPLASH = "has_splash"
    ON_PLAY = "on_play"
    NUM_ON_PLAY = "num_on_play"
    NUM_MULLIGANS = "num_mulligans"
    NUM_MULLIGANS_SUM = "num_mulligans_sum"
    OPP_NUM_MULLIGANS = "opp_num_mulligans"
    OPP_NUM_MULLIGANS_SUM = "opp_num_mulligans_sum"
    OPP_COLORS = "opp_colors"
    NUM_TURNS = "num_turns"
    NUM_TURNS_SUM = "num_turns_sum"
    WON = "won"
    NUM_WON = "num_won"
    OPENING_HAND = "opening_hand"
    WON_OPENING_HAND = "won_opening_hand"
    DRAWN = "drawn"
    WON_DRAWN = "won_drawn"
    TUTORED = "tutored"
    WON_TUTORED = "won_tutored"
    DECK = "deck"
    WON_DECK = "won_deck"
    SIDEBOARD = "sideboard"
    WON_SIDEBOARD = "won_sideboard"
    NUM_GNS = "num_gns"
    WON_NUM_GNS = "won_num_gns"
    # card
    SET_CODE = "set_code"
    COLOR = "color"
    RARITY = "rarity"
    COLOR_IDENTITY = "color_identity"
    CARD_TYPE = "card_type"
    SUBTYPE = "subtype"
    MANA_VALUE = "mana_value"
    DECK_MANA_VALUE = "deck_mana_value"
    DECK_LANDS = "deck_lands"
    DECK_SPELLS = "deck_spells"
    MANA_COST = "mana_cost"
    POWER = "power"
    TOUGHNESS = "toughness"
    IS_BONUS_SHEET = "is_bonus_sheet"
    IS_DFC = "is_dfc"
    ORACLE_TEXT = "oracle_text"
    CARD_JSON = "card_json"
    # agg extensions
    PICKED_MATCH_WR = "picked_match_wr"
    TROPHY_RATE = "trophy_rate"
    GAME_WR = "game_wr"
    ALSA = "alsa"
    ATA = "ata"
    NUM_GP = "num_gp"
    PCT_GP = "pct_gp"
    GP_WR = "gp_wr"
    NUM_OH = "num_oh"
    OH_WR = "oh_wr"
    NUM_GIH = "num_gih"
    NUM_GIH_WON = "num_gih_won"
    GIH_WR = "gih_wr"
    GNS_WR = "gns_wr"
    IWD = "iwd"
    NUM_IN_POOL = "num_in_pool"
    NUM_IN_POOL_TOTAL = "num_in_pool_total"
    IN_POOL_WR = "in_pool_wr"
    DECK_TOTAL = "deck_total"
    WON_DECK_TOTAL = "won_deck_total"
    GP_WR_MEAN = "gp_wr_mean"
    GP_WR_EXCESS = "gp_wr_excess"
    GP_WR_VAR = "gp_wr_var"
    GP_WR_STDEV = "gp_wr_stdev"
    GP_WR_Z = "gp_wr_z"
    GIH_TOTAL = "gih_total"
    WON_GIH_TOTAL = "won_gih_total"
    GIH_WR_MEAN = "gih_wr_mean"
    GIH_WR_EXCESS = "gih_wr_excess"
    GIH_WR_VAR = "gih_wr_var"
    GIH_WR_STDEV = "gh_wr_stdev"
    GIH_WR_Z = "gih_wr_z"
    DECK_MANA_VALUE_AVG = "deck_mana_value_avg"
    DECK_LANDS_AVG = "deck_lands_avg"
    DECK_SPELLS_AVG = "deck_spells_avg"



from dataclasses import dataclass
from collections.abc import Callable

import polars as pl

from spells.enums import View, ColName, ColType


@dataclass(frozen=True)
class ColSpec:
    col_type: ColType
    expr: pl.Expr | Callable[..., pl.Expr] | None = None
    views: list[View] | None = None
    version: str | None = None


@dataclass(frozen=True)
class ColDef:
    name: str
    col_type: ColType
    expr: pl.Expr | tuple[pl.Expr, ...]
    views: set[View]
    dependencies: set[str]
    signature: str


default_columns = [
    ColName.COLOR,
    ColName.RARITY,
    ColName.NUM_SEEN,
    ColName.ALSA,
    ColName.NUM_TAKEN,
    ColName.ATA,
    ColName.NUM_GP,
    ColName.PCT_GP,
    ColName.GP_WR,
    ColName.NUM_OH,
    ColName.OH_WR,
    ColName.NUM_GIH,
    ColName.GIH_WR,
]

_specs: dict[str, ColSpec] = {
    ColName.NAME: ColSpec(
        col_type=ColType.GROUP_BY,
        views=[View.CARD],
    ),
    ColName.EXPANSION: ColSpec(
        col_type=ColType.GROUP_BY,
        views=[View.GAME, View.DRAFT],
    ),
    ColName.EVENT_TYPE: ColSpec(
        col_type=ColType.GROUP_BY,
        views=[View.GAME, View.DRAFT],
    ),
    ColName.DRAFT_ID: ColSpec(
        views=[View.GAME, View.DRAFT],
        col_type=ColType.FILTER_ONLY,
    ),
    ColName.DRAFT_TIME: ColSpec(
        col_type=ColType.FILTER_ONLY,
        views=[View.GAME, View.DRAFT],
    ),
    ColName.DRAFT_DATE: ColSpec(
        col_type=ColType.GROUP_BY,
        expr=pl.col(ColName.DRAFT_TIME).str.to_datetime("%Y-%m-%d %H:%M:%S").dt.date(),
    ),
    ColName.FORMAT_DAY: ColSpec(
        col_type=ColType.GROUP_BY,
        expr=lambda set_context: (
            pl.col(ColName.DRAFT_DATE)
            - pl.lit(set_context["release_time"])
            .str.to_datetime("%Y-%m-%d %H:%M:%S")
            .dt.date()
        ).dt.total_days()
        + 1,
    ),
    ColName.DRAFT_DAY_OF_WEEK: ColSpec(
        col_type=ColType.GROUP_BY,
        expr=pl.col(ColName.DRAFT_TIME)
        .str.to_datetime("%Y-%m-%d %H:%M:%S")
        .dt.weekday(),
    ),
    ColName.DRAFT_HOUR: ColSpec(
        col_type=ColType.GROUP_BY,
        expr=pl.col(ColName.DRAFT_TIME).str.to_datetime("%Y-%m-%d %H:%M:%S").dt.hour(),
    ),
    ColName.DRAFT_WEEK: ColSpec(
        col_type=ColType.GROUP_BY,
        expr=pl.col(ColName.DRAFT_TIME).str.to_datetime("%Y-%m-%d %H:%M:%S").dt.week(),
    ),
    ColName.FORMAT_WEEK: ColSpec(
        col_type=ColType.GROUP_BY, expr=(pl.col(ColName.FORMAT_DAY) - 1) // 7 + 1
    ),
    ColName.RANK: ColSpec(
        col_type=ColType.GROUP_BY,
        views=[View.GAME, View.DRAFT],
    ),
    ColName.USER_N_GAMES_BUCKET: ColSpec(
        col_type=ColType.GROUP_BY,
        views=[View.DRAFT, View.GAME],
    ),
    ColName.USER_GAME_WIN_RATE_BUCKET: ColSpec(
        col_type=ColType.GROUP_BY,
        views=[View.DRAFT, View.GAME],
    ),
    ColName.PLAYER_COHORT: ColSpec(
        col_type=ColType.GROUP_BY,
        expr=pl.when(pl.col(ColName.USER_N_GAMES_BUCKET) < 100)
        .then(pl.lit("Other"))
        .otherwise(
            pl.when(pl.col(ColName.USER_GAME_WIN_RATE_BUCKET) > 0.57)
            .then(pl.lit("Top"))
            .otherwise(
                pl.when(pl.col(ColName.USER_GAME_WIN_RATE_BUCKET) < 0.49)
                .then(pl.lit("Bottom"))
                .otherwise(pl.lit("Middle"))
            )
        ),
    ),
    ColName.EVENT_MATCH_WINS: ColSpec(
        col_type=ColType.GROUP_BY,
        views=[View.DRAFT],
    ),
    ColName.EVENT_MATCH_WINS_SUM: ColSpec(
        col_type=ColType.PICK_SUM,
        views=[View.DRAFT],
        expr=pl.col(ColName.EVENT_MATCH_WINS),
    ),
    ColName.EVENT_MATCH_LOSSES: ColSpec(
        col_type=ColType.GROUP_BY,
        views=[View.DRAFT],
    ),
    ColName.EVENT_MATCH_LOSSES_SUM: ColSpec(
        col_type=ColType.PICK_SUM,
        expr=pl.col(ColName.EVENT_MATCH_LOSSES),
    ),
    ColName.EVENT_MATCHES: ColSpec(
        col_type=ColType.GROUP_BY,
        expr=pl.col(ColName.EVENT_MATCH_WINS) + pl.col(ColName.EVENT_MATCH_LOSSES),
    ),
    ColName.EVENT_MATCHES_SUM: ColSpec(
        col_type=ColType.PICK_SUM,
        expr=pl.col(ColName.EVENT_MATCHES),
    ),
    ColName.IS_TROPHY: ColSpec(
        col_type=ColType.GROUP_BY,
        expr=pl.when(pl.col(ColName.EVENT_TYPE) == "Traditional")
        .then(pl.col(ColName.EVENT_MATCH_WINS) == 3)
        .otherwise(pl.col(ColName.EVENT_MATCH_WINS) == 7),
    ),
    ColName.IS_TROPHY_SUM: ColSpec(
        col_type=ColType.PICK_SUM,
        expr=pl.col(ColName.IS_TROPHY),
    ),
    ColName.PACK_NUMBER: ColSpec(
        col_type=ColType.FILTER_ONLY,  # use pack_num
        views=[View.DRAFT],
    ),
    ColName.PACK_NUM: ColSpec(
        col_type=ColType.GROUP_BY,
        expr=pl.col(ColName.PACK_NUMBER) + 1,
    ),
    ColName.PICK_NUMBER: ColSpec(
        col_type=ColType.FILTER_ONLY,  # use pick_num
        views=[View.DRAFT],
    ),
    ColName.PICK_NUM: ColSpec(
        col_type=ColType.GROUP_BY,
        expr=pl.col(ColName.PICK_NUMBER) + 1,
    ),
    ColName.TAKEN_AT: ColSpec(
        col_type=ColType.PICK_SUM,
        expr=pl.col(ColName.PICK_NUM),
    ),
    ColName.NUM_TAKEN: ColSpec(
        col_type=ColType.PICK_SUM,
        expr=pl.when(pl.col(ColName.PICK).is_not_null()).then(1).otherwise(0),
    ),
    ColName.NUM_DRAFTS: ColSpec(
        col_type=ColType.PICK_SUM,
        expr=pl.when(
            (pl.col(ColName.PACK_NUMBER) == 0) & (pl.col(ColName.PICK_NUMBER) == 0)
        )
        .then(1)
        .otherwise(0),
    ),
    ColName.PICK: ColSpec(
        col_type=ColType.FILTER_ONLY,
        views=[View.DRAFT],
    ),
    ColName.PICK_MAINDECK_RATE: ColSpec(
        col_type=ColType.PICK_SUM,
        views=[View.DRAFT],
    ),
    ColName.PICK_SIDEBOARD_IN_RATE: ColSpec(
        col_type=ColType.PICK_SUM,
        views=[View.DRAFT],
    ),
    ColName.PACK_CARD: ColSpec(
        col_type=ColType.NAME_SUM,
        views=[View.DRAFT],
    ),
    ColName.LAST_SEEN: ColSpec(
        col_type=ColType.NAME_SUM,
        expr=lambda name: pl.col(f"pack_card_{name}")
        * pl.min_horizontal(ColName.PICK_NUM, 8),
    ),
    ColName.NUM_SEEN: ColSpec(
        col_type=ColType.NAME_SUM,
        expr=lambda name: pl.col(f"pack_card_{name}") * (pl.col(ColName.PICK_NUM) <= 8),
    ),
    ColName.POOL: ColSpec(
        col_type=ColType.NAME_SUM,
        views=[View.DRAFT],
    ),
    ColName.GAME_TIME: ColSpec(
        col_type=ColType.FILTER_ONLY,
        views=[View.GAME],
    ),
    ColName.GAME_DATE: ColSpec(
        col_type=ColType.GROUP_BY,
        expr=pl.col(ColName.GAME_TIME).str.to_datetime("%Y-%m-%d %H-%M-%S").dt.date(),
    ),
    ColName.GAME_DAY_OF_WEEK: ColSpec(
        col_type=ColType.GROUP_BY,
        expr=pl.col(ColName.GAME_TIME)
        .str.to_datetime("%Y-%m-%d %H-%M-%S")
        .dt.weekday(),
    ),
    ColName.GAME_HOUR: ColSpec(
        col_type=ColType.GROUP_BY,
        expr=pl.col(ColName.GAME_TIME).str.to_datetime("%Y-%m-%d %H-%M-%S").dt.hour(),
    ),
    ColName.GAME_WEEK: ColSpec(
        col_type=ColType.GROUP_BY,
        expr=pl.col(ColName.GAME_TIME).str.to_datetime("%Y-%m-%d %H-%M-%S").dt.week(),
    ),
    ColName.BUILD_INDEX: ColSpec(
        col_type=ColType.GROUP_BY,
        views=[View.GAME],
    ),
    ColName.MATCH_NUMBER: ColSpec(
        col_type=ColType.GROUP_BY,
        views=[View.GAME],
    ),
    ColName.GAME_NUMBER: ColSpec(
        col_type=ColType.GROUP_BY,
        views=[View.GAME],
    ),
    ColName.NUM_GAMES: ColSpec(
        col_type=ColType.GAME_SUM,
        expr=pl.col(ColName.GAME_NUMBER).is_not_null(),
    ),
    ColName.NUM_MATCHES: ColSpec(
        col_type=ColType.GAME_SUM,
        expr=pl.col(ColName.GAME_NUMBER) == 1,
    ),
    ColName.NUM_EVENTS: ColSpec(
        col_type=ColType.GAME_SUM,
        expr=(pl.col(ColName.GAME_NUMBER) == 1) & (pl.col(ColName.MATCH_NUMBER) == 1),
    ),
    ColName.OPP_RANK: ColSpec(
        col_type=ColType.GROUP_BY,
        views=[View.GAME],
    ),
    ColName.MAIN_COLORS: ColSpec(
        col_type=ColType.GROUP_BY,
        views=[View.GAME],
    ),
    ColName.NUM_COLORS: ColSpec(
        col_type=ColType.GROUP_BY,
        expr=pl.col(ColName.MAIN_COLORS).str.len_chars(),
    ),
    ColName.SPLASH_COLORS: ColSpec(
        col_type=ColType.GROUP_BY,
        views=[View.GAME],
    ),
    ColName.HAS_SPLASH: ColSpec(
        col_type=ColType.GROUP_BY,
        expr=pl.col(ColName.SPLASH_COLORS).str.len_chars() > 0,
    ),
    ColName.ON_PLAY: ColSpec(
        col_type=ColType.GROUP_BY,
        views=[View.GAME],
    ),
    ColName.NUM_ON_PLAY: ColSpec(
        col_type=ColType.GAME_SUM,
        expr=pl.col(ColName.ON_PLAY),
    ),
    ColName.NUM_MULLIGANS: ColSpec(
        col_type=ColType.GROUP_BY,
        views=[View.GAME],
    ),
    ColName.NUM_MULLIGANS_SUM: ColSpec(
        col_type=ColType.GAME_SUM,
        expr=pl.col(ColName.NUM_MULLIGANS),
    ),
    ColName.OPP_NUM_MULLIGANS: ColSpec(
        col_type=ColType.GAME_SUM,
        views=[View.GAME],
    ),
    ColName.OPP_NUM_MULLIGANS_SUM: ColSpec(
        col_type=ColType.GAME_SUM,
        expr=pl.col(ColName.OPP_NUM_MULLIGANS),
    ),
    ColName.OPP_COLORS: ColSpec(
        col_type=ColType.GROUP_BY,
        views=[View.GAME],
    ),
    ColName.NUM_TURNS: ColSpec(
        col_type=ColType.GROUP_BY,
        views=[View.GAME],
    ),
    ColName.NUM_TURNS_SUM: ColSpec(
        col_type=ColType.GAME_SUM,
        expr=pl.col(ColName.NUM_TURNS),
    ),
    ColName.WON: ColSpec(
        col_type=ColType.GROUP_BY,
        views=[View.GAME],
    ),
    ColName.NUM_WON: ColSpec(
        col_type=ColType.GAME_SUM,
        expr=pl.col(ColName.WON),
    ),
    ColName.OPENING_HAND: ColSpec(
        col_type=ColType.NAME_SUM,
        views=[View.GAME],
    ),
    ColName.WON_OPENING_HAND: ColSpec(
        col_type=ColType.NAME_SUM,
        expr=lambda name: pl.col(f"opening_hand_{name}") * pl.col(ColName.WON),
    ),
    ColName.DRAWN: ColSpec(
        col_type=ColType.NAME_SUM,
        views=[View.GAME],
    ),
    ColName.WON_DRAWN: ColSpec(
        col_type=ColType.NAME_SUM,
        expr=lambda name: pl.col(f"drawn_{name}") * pl.col(ColName.WON),
    ),
    ColName.TUTORED: ColSpec(
        col_type=ColType.NAME_SUM,
        views=[View.GAME],
    ),
    ColName.WON_TUTORED: ColSpec(
        col_type=ColType.NAME_SUM,
        expr=lambda name: pl.col(f"tutored_{name}") * pl.col(ColName.WON),
    ),
    ColName.DECK: ColSpec(
        col_type=ColType.NAME_SUM,
        views=[View.GAME],
    ),
    ColName.WON_DECK: ColSpec(
        col_type=ColType.NAME_SUM,
        expr=lambda name: pl.col(f"deck_{name}") * pl.col(ColName.WON),
    ),
    ColName.SIDEBOARD: ColSpec(
        col_type=ColType.NAME_SUM,
        views=[View.GAME],
    ),
    ColName.WON_SIDEBOARD: ColSpec(
        col_type=ColType.NAME_SUM,
        expr=lambda name: pl.col(f"sideboard_{name}") * pl.col(ColName.WON),
    ),
    ColName.NUM_GNS: ColSpec(
        col_type=ColType.NAME_SUM,
        expr=lambda name: pl.max_horizontal(
            0,
            pl.col(f"deck_{name}")
            - pl.col(f"drawn_{name}")
            - pl.col(f"tutored_{name}")
            - pl.col(f"opening_hand_{name}"),
        ),
    ),
    ColName.WON_NUM_GNS: ColSpec(
        col_type=ColType.NAME_SUM,
        expr=lambda name: pl.col(ColName.WON) * pl.col(f"num_gns_{name}"),
    ),
    ColName.SET_CODE: ColSpec(
        col_type=ColType.CARD_ATTR,
    ),
    ColName.COLOR: ColSpec(
        col_type=ColType.CARD_ATTR,
    ),
    ColName.RARITY: ColSpec(
        col_type=ColType.CARD_ATTR,
    ),
    ColName.COLOR_IDENTITY: ColSpec(
        col_type=ColType.CARD_ATTR,
    ),
    ColName.CARD_TYPE: ColSpec(
        col_type=ColType.CARD_ATTR,
    ),
    ColName.SUBTYPE: ColSpec(
        col_type=ColType.CARD_ATTR,
    ),
    ColName.MANA_VALUE: ColSpec(
        col_type=ColType.CARD_ATTR,
    ),
    ColName.DECK_MANA_VALUE: ColSpec(
        col_type=ColType.NAME_SUM,
        expr=lambda name, card_context: card_context[name][ColName.MANA_VALUE]
        * pl.col(f"deck_{name}"),
    ),
    ColName.DECK_LANDS: ColSpec(
        col_type=ColType.NAME_SUM,
        expr=lambda name, card_context: pl.col(f"deck_{name}")
        * (1 if "Land" in card_context[name][ColName.CARD_TYPE] else 0),
    ),
    ColName.DECK_SPELLS: ColSpec(
        col_type=ColType.NAME_SUM,
        expr=lambda name: pl.col(f"deck_{name}") - pl.col(f"deck_lands_{name}"),
    ),
    ColName.MANA_COST: ColSpec(
        col_type=ColType.CARD_ATTR,
    ),
    ColName.POWER: ColSpec(
        col_type=ColType.CARD_ATTR,
    ),
    ColName.TOUGHNESS: ColSpec(
        col_type=ColType.CARD_ATTR,
    ),
    ColName.IS_BONUS_SHEET: ColSpec(
        col_type=ColType.CARD_ATTR,
    ),
    ColName.IS_DFC: ColSpec(
        col_type=ColType.CARD_ATTR,
    ),
    ColName.ORACLE_TEXT: ColSpec(
        col_type=ColType.CARD_ATTR,
    ),
    ColName.CARD_JSON: ColSpec(
        col_type=ColType.CARD_ATTR,
    ),
    ColName.PICKED_MATCH_WR: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.EVENT_MATCH_WINS_SUM) / pl.col(ColName.EVENT_MATCHES_SUM),
    ),
    ColName.TROPHY_RATE: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.IS_TROPHY_SUM) / pl.col(ColName.NUM_TAKEN),
    ),
    ColName.GAME_WR: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.NUM_WON) / pl.col(ColName.NUM_GAMES),
    ),
    ColName.ALSA: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.LAST_SEEN) / pl.col(ColName.NUM_SEEN),
    ),
    ColName.ATA: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.TAKEN_AT) / pl.col(ColName.NUM_TAKEN),
    ),
    ColName.NUM_GP: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.DECK),
    ),
    ColName.PCT_GP: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.DECK) / (pl.col(ColName.DECK) + pl.col(ColName.SIDEBOARD)),
    ),
    ColName.GP_WR: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.WON_DECK) / pl.col(ColName.DECK),
    ),
    ColName.NUM_OH: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.OPENING_HAND),
    ),
    ColName.OH_WR: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.WON_OPENING_HAND) / pl.col(ColName.OPENING_HAND),
    ),
    ColName.NUM_GIH: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.OPENING_HAND) + pl.col(ColName.DRAWN),
    ),
    ColName.NUM_GIH_WON: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.WON_OPENING_HAND) + pl.col(ColName.WON_DRAWN),
    ),
    ColName.GIH_WR: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.NUM_GIH_WON) / pl.col(ColName.NUM_GIH),
    ),
    ColName.GNS_WR: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.WON_NUM_GNS) / pl.col(ColName.NUM_GNS),
    ),
    ColName.IWD: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.GIH_WR) - pl.col(ColName.GNS_WR),
    ),
    ColName.NUM_IN_POOL: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.DECK) + pl.col(ColName.SIDEBOARD),
    ),
    ColName.NUM_IN_POOL_TOTAL: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.NUM_IN_POOL).sum(),
    ),
    ColName.IN_POOL_WR: ColSpec(
        col_type=ColType.AGG,
        expr=(pl.col(ColName.WON_DECK) + pl.col(ColName.WON_SIDEBOARD))
        / pl.col(ColName.NUM_IN_POOL),
    ),
    ColName.DECK_TOTAL: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.DECK).sum(),
    ),
    ColName.WON_DECK_TOTAL: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.WON_DECK).sum(),
    ),
    ColName.GP_WR_MEAN: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.WON_DECK_TOTAL) / pl.col(ColName.DECK_TOTAL),
    ),
    ColName.GP_WR_EXCESS: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.GP_WR) - pl.col(ColName.GP_WR_MEAN),
    ),
    ColName.GP_WR_VAR: ColSpec(
        col_type=ColType.AGG,
        expr=(pl.col(ColName.GP_WR_EXCESS).pow(2) * pl.col(ColName.NUM_GP)).sum()
        / pl.col(ColName.DECK_TOTAL),
    ),
    ColName.GP_WR_STDEV: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.GP_WR_VAR).sqrt(),
    ),
    ColName.GP_WR_Z: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.GP_WR_EXCESS) / pl.col(ColName.GP_WR_STDEV),
    ),
    ColName.GIH_TOTAL: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.NUM_GIH).sum(),
    ),
    ColName.WON_GIH_TOTAL: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.NUM_GIH_WON).sum(),
    ),
    ColName.GIH_WR_MEAN: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.WON_GIH_TOTAL) / pl.col(ColName.GIH_TOTAL),
    ),
    ColName.GIH_WR_EXCESS: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.GIH_WR) - pl.col(ColName.GIH_WR_MEAN),
    ),
    ColName.GIH_WR_VAR: ColSpec(
        col_type=ColType.AGG,
        expr=(pl.col(ColName.GIH_WR_EXCESS).pow(2) * pl.col(ColName.NUM_GIH)).sum()
        / pl.col(ColName.GIH_TOTAL),
    ),
    ColName.GIH_WR_STDEV: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.GIH_WR_VAR).sqrt(),
    ),
    ColName.GIH_WR_Z: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.GIH_WR_EXCESS) / pl.col(ColName.GIH_WR_STDEV),
    ),
    ColName.DECK_MANA_VALUE_AVG: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.DECK_MANA_VALUE) / pl.col(ColName.DECK_SPELLS),
    ),
    ColName.DECK_LANDS_AVG: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.DECK_LANDS) / pl.col(ColName.NUM_GAMES),
    ),
    ColName.DECK_SPELLS_AVG: ColSpec(
        col_type=ColType.AGG,
        expr=pl.col(ColName.DECK_SPELLS) / pl.col(ColName.NUM_GAMES),
    ),
}

for item in ColName:
    assert item in _specs, f"column {item} enumerated but not specified"


class GetSpecs:
    def __init__(self, spec_dict: dict[str, ColSpec]):
        self._specs = spec_dict

    def __call__(self):
        return dict(self._specs)


get_specs = GetSpecs(_specs)



import math

import polars as pl

from spells.enums import ColType, ColName
from spells.columns import ColSpec
from spells.cache import spells_print


def print_ext(ext: dict[str, ColSpec]) -> None:
    spells_print("create", "Created extensions:")
    for key in ext:
        print("\t" + key)


def context_cols(attr, silent: bool = False) -> dict[str, ColSpec]:
    ext = {
        f"seen_{attr}": ColSpec(
            col_type=ColType.NAME_SUM,
            expr=(
                lambda name, card_context: pl.lit(None)
                if card_context[name][attr] is None
                or math.isnan(card_context[name][attr])
                else pl.when(pl.col(f"pack_card_{name}") > 0)
                .then(card_context[name][attr])
                .otherwise(None)
            ),
        ),
        f"pick_{attr}": ColSpec(
            col_type=ColType.PICK_SUM,
            expr=lambda name, card_context: pl.lit(None)
            if card_context[name][attr] is None or math.isnan(card_context[name][attr])
            else card_context[name][attr],
        ),
        f"seen_{attr}_greater": ColSpec(
            col_type=ColType.NAME_SUM,
            expr=lambda name: pl.col(f"seen_{attr}_{name}") > pl.col(f"pick_{attr}"),
        ),
        f"seen_{attr}_less": ColSpec(
            col_type=ColType.NAME_SUM,
            expr=lambda name: pl.col(f"seen_{attr}_{name}") < pl.col(f"pick_{attr}"),
        ),
        f"greatest_{attr}_seen": ColSpec(
            col_type=ColType.PICK_SUM,
            expr=lambda names: pl.max_horizontal(
                [pl.col(f"seen_{attr}_{name}") for name in names]
            ),
        ),
        f"least_{attr}_seen": ColSpec(
            col_type=ColType.PICK_SUM,
            expr=lambda names: pl.min_horizontal(
                [pl.col(f"seen_{attr}_{name}") for name in names]
            ),
        ),
        f"pick_{attr}_rank_greatest": ColSpec(
            col_type=ColType.GROUP_BY,
            expr=lambda names: pl.sum_horizontal(
                [pl.col(f"seen_{attr}_greater_{name}") for name in names]
            )
            + 1,
        ),
        f"pick_{attr}_rank_least": ColSpec(
            col_type=ColType.GROUP_BY,
            expr=lambda names: pl.sum_horizontal(
                [pl.col(f"seen_{attr}_less_{name}") for name in names]
            )
            + 1,
        ),
        f"pick_{attr}_rank_greatest_sum": ColSpec(
            col_type=ColType.PICK_SUM, expr=pl.col(f"pick_{attr}_rank_greatest")
        ),
        f"pick_{attr}_rank_least_sum": ColSpec(
            col_type=ColType.PICK_SUM, expr=pl.col(f"pick_{attr}_rank_least")
        ),
        f"pick_{attr}_vs_least": ColSpec(
            col_type=ColType.PICK_SUM,
            expr=pl.col(f"pick_{attr}") - pl.col(f"least_{attr}_seen"),
        ),
        f"pick_{attr}_vs_greatest": ColSpec(
            col_type=ColType.PICK_SUM,
            expr=pl.col(f"pick_{attr}") - pl.col(f"greatest_{attr}_seen"),
        ),
        f"pick_{attr}_vs_least_mean": ColSpec(
            col_type=ColType.AGG,
            expr=pl.col(f"pick_{attr}_vs_least") / pl.col(ColName.NUM_TAKEN),
        ),
        f"pick_{attr}_vs_greatest_mean": ColSpec(
            col_type=ColType.AGG,
            expr=pl.col(f"pick_{attr}_vs_greatest") / pl.col(ColName.NUM_TAKEN),
        ),
        f"least_{attr}_taken": ColSpec(
            col_type=ColType.PICK_SUM,
            expr=pl.col(f"pick_{attr}") <= pl.col(f"least_{attr}_seen"),
        ),
        f"least_{attr}_taken_rate": ColSpec(
            col_type=ColType.AGG,
            expr=pl.col(f"least_{attr}_taken") / pl.col(ColName.NUM_TAKEN),
        ),
        f"greatest_{attr}_taken": ColSpec(
            col_type=ColType.PICK_SUM,
            expr=pl.col(f"pick_{attr}") >= pl.col(f"greatest_{attr}_seen"),
        ),
        f"greatest_{attr}_taken_rate": ColSpec(
            col_type=ColType.AGG,
            expr=pl.col(f"greatest_{attr}_taken") / pl.col(ColName.NUM_TAKEN),
        ),
        f"pick_{attr}_mean": ColSpec(
            col_type=ColType.AGG,
            expr=pl.col(f"pick_{attr}") / pl.col(ColName.NUM_TAKEN),
        ),
    }

    if not silent:
        print_ext(ext)

    return ext


def stat_cols(attr: str, silent: bool = False) -> dict[str, ColSpec]:
    ext = {
        f"{attr}_deck_weight_group": ColSpec(
            col_type=ColType.AGG, expr=pl.col(f"{attr}") * pl.col(ColName.DECK)
        ),
        f"{attr}_deck_weight_total": ColSpec(
            col_type=ColType.AGG, expr=pl.col(f"{attr}_deck_weight_group").sum()
        ),
        f"{attr}_dw_mean": ColSpec(
            col_type=ColType.AGG,
            expr=pl.col(f"{attr}_deck_weight_total") / pl.col(ColName.DECK_TOTAL),
        ),
        f"{attr}_dw_excess": ColSpec(
            col_type=ColType.AGG,
            expr=pl.col(f"{attr}") - pl.col(f"{attr}_dw_mean"),
        ),
        f"{attr}_dw_var": ColSpec(
            col_type=ColType.AGG,
            expr=(pl.col(f"{attr}_dw_excess").pow(2) * pl.col(ColName.DECK)).sum()
            / pl.col(ColName.DECK_TOTAL),
        ),
        f"{attr}_dw_stdev": ColSpec(
            col_type=ColType.AGG,
            expr=pl.col(f"{attr}_dw_var").sqrt(),
        ),
        f"{attr}_dwz": ColSpec(
            col_type=ColType.AGG,
            expr=pl.col(f"{attr}_dw_excess") / pl.col(f"{attr}_dw_stdev"),
        ),
        f"{attr}_pool_weight_group": ColSpec(
            col_type=ColType.AGG, expr=pl.col(f"{attr}") * pl.col(ColName.NUM_IN_POOL)
        ),
        f"{attr}_pool_weight_total": ColSpec(
            col_type=ColType.AGG, expr=pl.col(f"{attr}_pool_weight_group").sum()
        ),
        f"{attr}_pw_mean": ColSpec(
            col_type=ColType.AGG,
            expr=pl.col(f"{attr}_pool_weight_total")
            / pl.col(ColName.NUM_IN_POOL_TOTAL),
        ),
        f"{attr}_pw_excess": ColSpec(
            col_type=ColType.AGG,
            expr=pl.col(f"{attr}") - pl.col(f"{attr}_pw_mean"),
        ),
        f"{attr}_pw_var": ColSpec(
            col_type=ColType.AGG,
            expr=(
                pl.col(f"{attr}_pw_excess").pow(2) * pl.col(ColName.NUM_IN_POOL)
            ).sum()
            / pl.col(ColName.NUM_IN_POOL_TOTAL),
        ),
        f"{attr}_pw_stdev": ColSpec(
            col_type=ColType.AGG,
            expr=pl.col(f"{attr}_pw_var").sqrt(),
        ),
        f"{attr}_pwz": ColSpec(
            col_type=ColType.AGG,
            expr=pl.col(f"{attr}_pw_excess") / pl.col(f"{attr}_pw_stdev"),
        ),
    }

    if not silent:
        print_ext(ext)

    return ext



from dataclasses import dataclass

import spells.columns
import spells.filter
from spells.enums import View, ColName, ColType
from spells.columns import ColDef


@dataclass(frozen=True)
class Manifest:
    columns: tuple[str, ...]
    col_def_map: dict[str, ColDef]
    base_view_group_by: frozenset[str]
    view_cols: dict[View, frozenset[str]]
    group_by: tuple[str, ...]
    filter: spells.filter.Filter | None

    def __post_init__(self):
        # No name filter check
        if self.filter is not None:
            assert (
                "name" not in self.filter.lhs
            ), "Don't filter on 'name', include 'name' in groupbys and filter the final result instead"

        # Col in col_def_map check
        for col in self.columns:
            assert col in self.col_def_map, f"Undefined column {col}!"
            assert (
                self.col_def_map[col].col_type != ColType.GROUP_BY
            ), f"group_by column {col} must be passed as group_by"
            assert (
                self.col_def_map[col].col_type != ColType.FILTER_ONLY
            ), f"filter_only column {col} cannot be summoned"

        # base_view_groupbys have col_type GROUP_BY check
        for col in self.base_view_group_by:
            assert (
                self.col_def_map[col].col_type == ColType.GROUP_BY
            ), f"Invalid groupby {col}!"

        for view, cols_for_view in self.view_cols.items():
            for col in cols_for_view:
                # game sum cols on in game, and no NAME groupby
                assert self.col_def_map[col].col_type != ColType.GAME_SUM or (
                    view == View.GAME and ColName.NAME not in self.base_view_group_by
                ), f"Invalid manifest for GAME_SUM column {col}"
            if view != View.CARD:
                for col in self.base_view_group_by:
                    # base_view_groupbys in view_cols for view
                    assert (
                        col == ColName.NAME or col in cols_for_view
                    ), f"Groupby {col} not in view_cols[view]"
                # filter cols are in both base_views check
                if self.filter is not None:
                    for col in self.filter.lhs:
                        assert (
                            col in cols_for_view
                        ), f"filter col {col} not found in base view"

            if view == View.CARD:
                # name in groupbys check
                assert (
                    ColName.NAME in self.base_view_group_by
                ), "base views must groupby by name to join card attrs"

    def test_str(self):
        result = "{\n" + 2 * " " + "columns:\n"
        for c in sorted(self.columns):
            result += 4 * " " + c + "\n"
        result += 2 * " " + "base_view_group_by:\n"
        for c in sorted(self.base_view_group_by):
            result += 4 * " " + c + "\n"
        result += 2 * " " + "view_cols:\n"
        for v, view_cols in sorted(self.view_cols.items()):
            result += 4 * " " + v + ":\n"
            for c in sorted(view_cols):
                result += 6 * " " + c + "\n"
        result += 2 * " " + "group_by:\n"
        for c in sorted(self.group_by):
            result += 4 * " " + c + "\n"
        result += "}\n"

        return result


def _resolve_view_cols(
    col_set: frozenset[str],
    col_def_map: dict[str, ColDef],
) -> dict[View, frozenset[str]]:
    """
    For each view ('game', 'draft', and 'card'), return the columns
    that must be present at the aggregation step. 'name' need not be
    included, and 'pick' will be added if needed.
    """
    MAX_DEPTH = 1000
    unresolved_cols = col_set
    view_resolution = {}

    iter_num = 0
    while unresolved_cols and iter_num < MAX_DEPTH:
        iter_num += 1
        next_cols = frozenset()
        for col in unresolved_cols:
            cdef = col_def_map[col]
            if cdef.col_type == ColType.PICK_SUM:
                view_resolution[View.DRAFT] = view_resolution.get(
                    View.DRAFT, frozenset()
                ).union({ColName.PICK})
            # now determine views and deps
            if cdef.col_type == ColType.CARD_ATTR:
                view_resolution[View.CARD] = view_resolution.get(
                    View.CARD, frozenset()
                ).union({col})
            elif cdef.views:
                for view in cdef.views:
                    view_resolution[view] = view_resolution.get(
                        view, frozenset()
                    ).union({col})
            else:
                if cdef.dependencies is None:
                    raise ValueError(
                        f"Invalid column def: {col} has neither views nor dependencies!"
                    )
                if cdef.col_type != ColType.AGG:
                    fully_resolved = True
                    col_views = frozenset({View.GAME, View.DRAFT, View.CARD})
                    for dep in cdef.dependencies:
                        dep_views = frozenset()
                        for view, view_cols in view_resolution.items():
                            if dep in view_cols:
                                dep_views = dep_views.union({view})
                        if not dep_views:
                            fully_resolved = False
                            next_cols = next_cols.union({dep})
                        else:
                            col_views = col_views.intersection(dep_views)
                    if fully_resolved:
                        assert len(
                            col_views
                        ), f"Column {col} can't be defined in any views!"
                        for view in col_views:
                            if view not in view_resolution:
                                print(cdef)
                                assert False, f"Something went wrong with col {col}"

                            view_resolution[view] = view_resolution[view].union({col})
                    else:
                        next_cols = next_cols.union({col})
                else:
                    for dep in cdef.dependencies:
                        next_cols = next_cols.union({dep})
        unresolved_cols = next_cols

    if iter_num >= MAX_DEPTH:
        raise ValueError("broken dependency chain in column spec, loop probable")

    return view_resolution


def create(
    col_def_map: dict[str, ColDef],
    columns: list[str] | None = None,
    group_by: list[str] | None = None,
    filter_spec: dict | None = None,
):
    gbs = (ColName.NAME,) if group_by is None else tuple(group_by)

    if columns is None:
        cols = tuple(spells.columns.default_columns)
        if ColName.NAME not in gbs:
            cols = tuple(
                col for col in cols if col not in (ColName.COLOR, ColName.RARITY)
            )
    else:
        cols = tuple(columns)

    m_filter = spells.filter.from_spec(filter_spec)

    col_set = frozenset(cols)
    col_set = col_set.union(frozenset(gbs) - {ColName.NAME})
    if m_filter is not None:
        col_set = col_set.union(m_filter.lhs)

    base_view_group_by = frozenset()

    for col in gbs:
        cdef = col_def_map[col]
        if cdef.col_type == ColType.GROUP_BY:
            base_view_group_by = base_view_group_by.union({col})
        elif cdef.col_type == ColType.CARD_ATTR:
            base_view_group_by = base_view_group_by.union({ColName.NAME})

    view_cols = _resolve_view_cols(col_set, col_def_map)

    needed_views = frozenset()
    if View.CARD in view_cols:
        needed_views = needed_views.union({View.CARD})

    draft_view_cols = view_cols.get(View.DRAFT, frozenset())
    game_view_cols = view_cols.get(View.GAME, frozenset())

    base_cols = draft_view_cols.union(game_view_cols)

    if base_cols == draft_view_cols:
        needed_views = needed_views.union({View.DRAFT})
    elif base_cols == game_view_cols:
        needed_views = needed_views.union({View.GAME})
    else:
        needed_views = needed_views.union({View.GAME, View.DRAFT})

    view_cols = {v: view_cols[v] for v in needed_views}

    return Manifest(
        columns=cols,
        col_def_map=col_def_map,
        base_view_group_by=base_view_group_by,
        view_cols=view_cols,
        group_by=gbs,
        filter=m_filter,
    )



"""
download public data sets from 17Lands.com and generate a card
file containing card attributes using MTGJSON

cli tool `spells`
"""

import functools
import gzip
import os
import re
import shutil
import sys
from enum import StrEnum

import wget
import polars as pl
from polars.exceptions import ComputeError

from spells import cards
from spells import cache
from spells.enums import View
from spells.schema import schema


DATASET_TEMPLATE = "{dataset_type}_data_public.{set_code}.{event_type}.csv.gz"
RESOURCE_TEMPLATE = (
    "https://17lands-public.s3.amazonaws.com/analysis_data/{dataset_type}_data/"
)


class FileFormat(StrEnum):
    CSV = "csv"
    PARQUET = "parquet"


class EventType(StrEnum):
    PREMIER = "PremierDraft"
    TRADITIONAL = "TradDraft"


# Fred Cirera via https://stackoverflow.com/questions/1094841/get-a-human-readable-version-of-a-file-size
def sizeof_fmt(num, suffix="B"):
    for unit in ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"):
        if abs(num) < 1024.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"


def cli() -> int:
    data_dir = cache.data_home()
    cache.spells_print("spells", f"[data home]={data_dir}")
    print()
    usage = """spells [add|refresh|remove|clean] [set_code]
            spells clean all
            spells info

    add: Download draft and game files from 17Lands.com and card file from MTGJSON.com and save to path 
        [data home]/external/[set code] (use $SPELLS_DATA_HOME or $XDG_DATA_HOME to configure). 
        Does not overwrite existing files. If any files are downloaded, existing local cache is cleared.
        [set code] should be the capitalized official set code for the draft release.

        e.g. $ spells add OTJ

    refresh: Force download and overwrite of existing files (for new data drops, use sparingly!). Clear 
        local cache.

    remove: Delete the [data home]/external/[set code] and [data home]/local/[set code] directories and their contents

    clean: Delete [data home]/local/[set code] data directory (your cache of aggregate parquet files), or all of them.

    info: No set code argument. Print info on all external and local files.
    """
    print_usage = functools.partial(cache.spells_print, "usage", usage)

    if len(sys.argv) < 2:
        print_usage()
        return 1

    mode = sys.argv[1]

    if mode == "info":
        return _info()

    if len(sys.argv) != 3:
        print_usage()
        return 1

    match mode:
        case "add":
            return _add(sys.argv[2])
        case "refresh":
            return _refresh(sys.argv[2])
        case "remove":
            return _remove(sys.argv[2])
        case "clean":
            return cache.clean(sys.argv[2])
        case _:
            print_usage()
            return 1


def _add(set_code: str, force_download=False):
    download_data_set(set_code, View.DRAFT, force_download=force_download)
    write_card_file(set_code, force_download=force_download)
    get_set_context(set_code)
    download_data_set(set_code, View.GAME, force_download=force_download)
    return 0


def _refresh(set_code: str):
    return _add(set_code, force_download=True)


def _remove(set_code: str):
    mode = "remove"
    dir_path = _external_set_path(set_code)
    if os.path.isdir(dir_path):
        with os.scandir(dir_path) as set_dir:
            count = 0
            for entry in set_dir:
                if not entry.name.endswith(".parquet"):
                    cache.spells_print(
                        mode,
                        f"Unexpected file {entry.name} found in external cache, please sort that out!",
                    )
                    return 1
                count += 1
                os.remove(entry)
            cache.spells_print(
                mode, f"Removed {count} files from external cache for set {set_code}"
            )
        os.rmdir(dir_path)
    else:
        cache.spells_print(mode, f"No external cache found for set {set_code}")

    return cache.clear(set_code)


def _info():
    mode = "info"
    external_path = cache.data_dir_path(cache.DataDir.EXTERNAL)

    suggest_add = set()
    suggest_remove = set()
    all_external = set()
    if os.path.isdir(external_path):
        cache.spells_print(mode, f"External archives found {external_path}")
        with os.scandir(external_path) as ext_dir:
            for entry in ext_dir:
                if entry.is_dir():
                    all_external.add(entry.name)
                    file_count = 0
                    cache.spells_print(mode, f"Archive {entry.name} contents:")
                    for item in os.scandir(entry):
                        if not re.match(f"^{entry.name}_.*\\.parquet", item.name):
                            print(
                                f"!!! imposter file {item.name}! Please sort that out"
                            )
                        print(f"    {item.name} {sizeof_fmt(os.stat(item).st_size)}")
                        file_count += 1
                    if file_count < 4:
                        suggest_add.add(entry.name)
                    if file_count > 4:
                        suggest_remove.add(entry.name)
                else:
                    cache.spells_print(
                        mode, f"Imposter file {entry.name}! Please sort that out"
                    )

    else:
        cache.spells_print(mode, "No external archives found")

    cache_path = cache.data_dir_path(cache.DataDir.CACHE)

    if os.path.isdir(cache_path):
        print()
        cache.spells_print(mode, f"Local cache found {cache_path}")
        with os.scandir(cache_path) as cache_dir:
            for entry in cache_dir:
                if entry.name not in all_external:
                    suggest_remove.add(entry.name)
                if entry.is_dir():
                    cache.spells_print(mode, f"Cache {entry.name} contents:")
                    parquet_num = 0
                    parquet_size = 0
                    for item in os.scandir(entry):
                        if item.name.endswith(".parquet"):
                            parquet_num += 1
                            parquet_size += os.stat(item).st_size
                        else:
                            print(
                                f"!!! imposter file {item.name}! Please sort that out"
                            )
                    print(f"    {parquet_num} cache files: {sizeof_fmt(parquet_size)}")
    else:
        print()
        cache.spells_print(mode, "No local cache found")

    print()
    for name in suggest_add:
        cache.spells_print(mode, f"Suggest `spells add {name}'")
    for name in suggest_remove:
        cache.spells_print(mode, f"Suggest `spells remove {name}'")

    return 0


def _external_set_path(set_code):
    return os.path.join(cache.data_dir_path(cache.DataDir.EXTERNAL), set_code)


def data_file_path(set_code, dataset_type: str, event_type=EventType.PREMIER):
    if dataset_type == "set_context":
        return os.path.join(_external_set_path(set_code), f"{set_code}_context.parquet")

    if dataset_type == "card":
        return os.path.join(_external_set_path(set_code), f"{set_code}_card.parquet")

    return os.path.join(
        _external_set_path(set_code), f"{set_code}_{event_type}_{dataset_type}.parquet"
    )


def _process_zipped_file(gzip_path, target_path):
    csv_path = gzip_path[:-3]
    # if polars supports streaming from file obj, we can just stream straight
    # from urllib.Request through GzipFile to sink_parquet without intermediate files
    with gzip.open(gzip_path, "rb") as f_in:
        with open(csv_path, "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)  # type: ignore

    os.remove(gzip_path)
    df = pl.scan_csv(csv_path, schema=schema(csv_path))
    try:
        df.sink_parquet(target_path)
    except ComputeError:
        df = pl.scan_csv(csv_path)
        cache.spells_print(
            "error",
            "Bad schema found, loading dataset into memory"
            + " and attempting to cast to correct schema",
        )
        select = [pl.col(name).cast(dtype) for name, dtype in schema(csv_path).items()]
        cast_df = df.select(select).collect()
        cast_df.write_parquet(target_path)

    os.remove(csv_path)


def download_data_set(
    set_code,
    dataset_type: View,
    event_type=EventType.PREMIER,
    force_download=False,
    clear_set_cache=True,
):
    mode = "refresh" if force_download else "add"
    cache.spells_print(mode, f"Downloading {dataset_type} dataset from 17Lands.com")

    if not os.path.isdir(set_dir := _external_set_path(set_code)):
        os.makedirs(set_dir)

    target_path = data_file_path(set_code, dataset_type)

    if os.path.isfile(target_path) and not force_download:
        cache.spells_print(
            mode,
            f"File {target_path} already exists, use `spells refresh {set_code}` to overwrite",
        )
        return 1

    dataset_file = DATASET_TEMPLATE.format(
        set_code=set_code, dataset_type=dataset_type, event_type=event_type
    )
    dataset_path = os.path.join(_external_set_path(set_code), dataset_file)
    wget.download(
        RESOURCE_TEMPLATE.format(dataset_type=dataset_type) + dataset_file,
        out=dataset_path,
    )
    print()

    cache.spells_print(
        mode, "Unzipping and transforming to parquet (this might take a few minutes)..."
    )
    _process_zipped_file(dataset_path, target_path)
    cache.spells_print(mode, f"Wrote file {target_path}")
    if clear_set_cache:
        cache.clear(set_code)

    return 0


def write_card_file(draft_set_code: str, force_download=False) -> int:
    """
    Write a csv containing basic information about draftable cards, such as rarity,
    set symbol, color, mana cost, and type.
    """
    mode = "refresh" if force_download else "add"

    cache.spells_print(
        mode, "Fetching card data from mtgjson.com and writing card file"
    )
    card_filepath = data_file_path(draft_set_code, View.CARD)
    if os.path.isfile(card_filepath) and not force_download:
        cache.spells_print(
            mode,
            f"File {card_filepath} already exists, use `spells refresh {draft_set_code}` to overwrite",
        )
        return 1

    draft_filepath = data_file_path(draft_set_code, View.DRAFT)

    if not os.path.isfile(draft_filepath):
        cache.spells_print(mode, f"Error: No draft file for set {draft_set_code}")
        return 1

    columns = pl.scan_parquet(draft_filepath).collect_schema().names()

    pattern = "^pack_card_"
    names = [
        re.split(pattern, name)[1]
        for name in columns
        if re.search(pattern, name) is not None
    ]

    card_df = cards.card_df(draft_set_code, names)

    card_df.write_parquet(card_filepath)

    cache.spells_print(mode, f"Wrote file {card_filepath}")
    return 0


def get_set_context(set_code: str, force_download=False) -> int:
    mode = "refresh" if force_download else "add"

    context_fp = data_file_path(set_code, "context")
    cache.spells_print(mode, "Calculating set context")
    if os.path.isfile(context_fp) and not force_download:
        cache.spells_print(
            mode,
            f"File {context_fp} already exists, use `spells refresh {set_code}` to overwrite",
        )
        return 1

    draft_fp = data_file_path(set_code, View.DRAFT)
    draft_view = pl.scan_parquet(draft_fp)

    context_df = draft_view.select(
        [
            pl.max("pick_number").alias("picks_per_pack") + 1,
            pl.min("draft_time").alias("release_time"),
        ]
    ).collect()

    context_df.write_parquet(context_fp)

    cache.spells_print(mode, f"Wrote file {context_fp}")
    return 0



"""
Test behavior of filters from dict specification
"""

import pytest
import polars as pl

import spells.filter

ROW_0 = {"int": 1, "float": 2.0, "text": "hi"}
ROW_1 = {"int": 0, "float": -0.4, "text": "foo"}
ROW_2 = {"int": -10, "float": 3.14, "text": "bar"}

TEST_DF = pl.DataFrame([ROW_0, ROW_1, ROW_2])


def format_test_string(test_string: str) -> str:
    """
    strip whitespace from each line to test pasted dataframe outputs
    """
    return "\n".join(
        [line.strip() for line in test_string.splitlines() if line.strip()]
    )


@pytest.mark.parametrize(
    "filter_spec, expected",
    [
        (None, None),
        (
            {"int": 1},
            """
shape: (1, 3)
┌─────┬───────┬──────┐
│ int ┆ float ┆ text │
│ --- ┆ ---   ┆ ---  │
│ i64 ┆ f64   ┆ str  │
╞═════╪═══════╪══════╡
│ 1   ┆ 2.0   ┆ hi   │
└─────┴───────┴──────┘
""",
        ),
        (
            {"lhs": "float", "rhs": 3, "op": "<"},
            """
shape: (2, 3)
┌─────┬───────┬──────┐
│ int ┆ float ┆ text │
│ --- ┆ ---   ┆ ---  │
│ i64 ┆ f64   ┆ str  │
╞═════╪═══════╪══════╡
│ 1   ┆ 2.0   ┆ hi   │
│ 0   ┆ -0.4  ┆ foo  │
└─────┴───────┴──────┘
        """,
        ),
        (
            {"$not": {"$or": [{"text": "foo"}, {"int": 1}]}},
            """
shape: (1, 3)
┌─────┬───────┬──────┐
│ int ┆ float ┆ text │
│ --- ┆ ---   ┆ ---  │
│ i64 ┆ f64   ┆ str  │
╞═════╪═══════╪══════╡
│ -10 ┆ 3.14  ┆ bar  │
└─────┴───────┴──────┘
        """,
        ),
        (
            {
                "$and": [
                    {"lhs": "text", "rhs": ["foo", "bar", "hi"], "op": "in"},
                    {"lhs": "int", "rhs": [1, 2], "op": "nin"},
                    {"lhs": "float", "rhs": 2.4, "op": "<"},
                ]
            },
            """
shape: (1, 3)
┌─────┬───────┬──────┐
│ int ┆ float ┆ text │
│ --- ┆ ---   ┆ ---  │
│ i64 ┆ f64   ┆ str  │
╞═════╪═══════╪══════╡
│ 0   ┆ -0.4  ┆ foo  │
└─────┴───────┴──────┘
            """,
        ),
    ],
)
def test_from_spec(filter_spec: dict | None, expected: str | None):
    test_filter = spells.filter.from_spec(filter_spec)

    if expected is None:
        assert test_filter is None
    else:
        assert test_filter is not None
        test_str = str(TEST_DF.filter(test_filter.expr))
        print(test_str)
        assert test_str == format_test_string(expected)


@pytest.mark.parametrize(
    "test_filter, expected",
    [
        (spells.filter.from_spec({"int": 1}), {"int"}),
        (
            spells.filter.from_spec({"$not": {"$or": [{"text": "foo"}, {"int": 1}]}}),
            {"text", "int"},
        ),
        (
            spells.filter.from_spec(
                {
                    "$and": [
                        {"lhs": "text", "rhs": ["foo", "bar", "hi"], "op": "in"},
                        {"lhs": "int", "rhs": [1, 2], "op": "nin"},
                        {"lhs": "float", "rhs": 2.4, "op": "<"},
                    ]
                }
            ),
            {"int", "float", "text"},
        ),
    ],
)
def test_lhs(test_filter, expected):
    assert test_filter.lhs == expected



"""
Module for caching the result of distributed dataframe calculations to parquet files.

Caches are keyed by a hash that is function of set code, aggregation type, base filter,
and groupbys.

Caches are cleared per-set when new files are downloaded.
"""

from enum import StrEnum
import os
import sys

import polars as pl


class DataDir(StrEnum):
    CACHE = "cache"
    EXTERNAL = "external"


def spells_print(mode, content):
    print(f"🪄 {mode} ✨ {content}")


def data_home() -> str:
    is_win = sys.platform == "win32"
    return os.path.expanduser(
        os.environ.get(
            "SPELLS_DATA_HOME",
            os.environ.get(
                "XDG_DATA_HOME",
                r"~\AppData\Local\Spells" if is_win else "~/.local/share/spells/",
            ),
        )
    )


def data_dir_path(cache_dir: DataDir) -> str:
    """
    Where 17Lands data is stored. MDU_DATA_DIR environment variable is used, if it exists,
    otherwise the cwd is used
    """
    is_win = sys.platform == "win32"

    ext = {
        DataDir.CACHE: "Cache" if is_win else "cache",
        DataDir.EXTERNAL: "External" if is_win else "external",
    }[cache_dir]

    data_dir = os.path.join(data_home(), ext)
    return data_dir


def cache_dir_for_set(set_code: str) -> str:
    return os.path.join(data_dir_path(DataDir.CACHE), set_code)


def cache_path_for_key(set_code: str, cache_key: str) -> str:
    return os.path.join(cache_dir_for_set(set_code), cache_key + ".parquet")


def cache_exists(set_code: str, cache_key: str) -> bool:
    return os.path.isdir(cache_dir_for_set(set_code)) and os.path.isfile(
        cache_path_for_key(set_code, cache_key)
    )


def read_cache(set_code: str, cache_key: str) -> pl.DataFrame:
    return pl.read_parquet(cache_path_for_key(set_code, cache_key))


def write_cache(set_code: str, cache_key: str, df: pl.DataFrame) -> None:
    cache_dir = cache_dir_for_set(set_code)
    if not os.path.isdir(cache_dir):
        os.makedirs(cache_dir)

    df.write_parquet(cache_path_for_key(set_code, cache_key))


def clean(set_code: str) -> int:
    mode = "clean"

    if set_code == "all":
        cache_dir = data_dir_path(DataDir.CACHE)
        with os.scandir(cache_dir) as set_dir:
            for entry in set_dir:
                clean(entry.name)
        return 0

    cache_dir = cache_dir_for_set(set_code)
    if os.path.isdir(cache_dir):
        with os.scandir(cache_dir) as set_dir:
            count = 0
            for entry in set_dir:
                if not entry.name.endswith(".parquet"):
                    spells_print(
                        mode,
                        f"Unexpected file {entry.name} found in local cache, please sort that out!",
                    )
                    return 1
                count += 1
                os.remove(entry)
            spells_print(
                mode, f"Removed {count} files from local cache for set {set_code}"
            )
        os.rmdir(cache_dir)
        spells_print(mode, f"Removed local cache dir {cache_dir}")
        return 0
    else:
        spells_print(mode, f"No local cache found for set {set_code}")
        return 0
